"""Defines a base curriculum class. Curricula are updated after every epoch."""

__all__ = [
    "Curriculum",
    "LinearCurriculum",
    "ConstantCurriculum",
    "EpisodeLengthCurriculum",
    "DistanceFromOriginCurriculum",
    "StepWhenSaturated",
]

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, TypeVar

import attrs
import jax
import jax.numpy as jnp
import xax
from jaxtyping import Array, PRNGKeyArray

from ksim.types import RewardState, Trajectory

T = TypeVar("T")


@jax.tree_util.register_dataclass
@dataclass(frozen=True)
class CurriculumState(Generic[T]):
    level: Array
    state: T


@attrs.define(frozen=True, kw_only=True)
class Curriculum(ABC, Generic[T]):
    """Base class for all curricula.

    Curricula should return a level between 0 and 1.
    """

    @abstractmethod
    def __call__(
        self,
        trajectory: Trajectory,
        rewards: RewardState,
        training_state: xax.State,
        prev_state: CurriculumState[T],
    ) -> CurriculumState[T]: ...

    @abstractmethod
    def get_initial_state(self, rng: PRNGKeyArray) -> CurriculumState[T]: ...


@attrs.define(frozen=True, kw_only=True)
class ConstantCurriculum(Curriculum[None]):
    """Constant curriculum."""

    level: float = attrs.field(default=1.0, validator=attrs.validators.ge(0.0))

    def __call__(
        self,
        trajectory: Trajectory,
        rewards: RewardState,
        training_state: xax.State,
        prev_state: CurriculumState[None],
    ) -> CurriculumState[None]:
        return prev_state

    def get_initial_state(self, rng: PRNGKeyArray) -> CurriculumState[None]:
        return CurriculumState(level=jnp.array(self.level, dtype=jnp.float32), state=None)


@attrs.define(frozen=True, kw_only=True)
class LinearCurriculum(Curriculum[None]):
    """Linear curriculum."""

    step_size: float = attrs.field(default=0.01, validator=attrs.validators.ge(0.0))
    step_every_n_epochs: int = attrs.field(default=1, validator=attrs.validators.ge(1))
    min_level: float = attrs.field(default=0.0, validator=attrs.validators.ge(0.0))

    def __call__(
        self,
        trajectory: Trajectory,
        rewards: RewardState,
        training_state: xax.State,
        prev_state: CurriculumState[None],
    ) -> CurriculumState[None]:
        level = (training_state.num_steps // self.step_every_n_epochs) * self.step_size
        level = jnp.clip(level, self.min_level, 1.0)
        level = jnp.full_like(prev_state.level, level)
        return CurriculumState(level=level, state=None)

    def get_initial_state(self, rng: PRNGKeyArray) -> CurriculumState[None]:
        return CurriculumState(level=jnp.array(self.min_level, dtype=jnp.float32), state=None)


@jax.tree_util.register_dataclass
@dataclass(frozen=True)
class EpisodeLengthCurriculumState:
    step_counter: Array
    ema_episode_length: Array


@attrs.define(frozen=True, kw_only=True)
class EpisodeLengthCurriculum(Curriculum[EpisodeLengthCurriculumState]):
    """Curriculum that updates the episode length."""

    num_levels: int = attrs.field(default=100, validator=attrs.validators.ge(1))
    increase_threshold: float = attrs.field(default=3.0, validator=attrs.validators.ge(0.0))
    decrease_threshold: float = attrs.field(default=3.0, validator=attrs.validators.ge(0.0))
    min_level_steps: int = attrs.field(default=1, validator=attrs.validators.ge(0))
    min_level: float = attrs.field(default=0.0, validator=attrs.validators.ge(0.0))
    ema_decay: float = attrs.field(default=0.9, validator=attrs.validators.ge(0.0))

    def __call__(
        self,
        trajectory: Trajectory,
        rewards: RewardState,
        training_state: xax.State,
        prev_state: CurriculumState[EpisodeLengthCurriculumState],
    ) -> CurriculumState[EpisodeLengthCurriculumState]:
        step_size = 1 / self.num_levels

        episode_length = trajectory.episode_length().mean()
        ema = self.ema_decay * prev_state.state.ema_episode_length + (1 - self.ema_decay) * episode_length

        steps = prev_state.state.step_counter
        level = prev_state.level
        can_step = steps == 0
        should_inc = (ema > self.increase_threshold) & can_step
        should_dec = (ema < self.decrease_threshold) & can_step
        next_steps = jnp.where(should_inc | should_dec, self.min_level_steps, jnp.maximum(steps - 1, 0))

        next_level = jnp.where(
            should_inc,
            level + step_size,
            jnp.where(should_dec, level - step_size, level),
        )
        next_level = jnp.clip(next_level, self.min_level, 1.0)
        next_steps = jnp.where(should_inc | should_dec, self.min_level_steps, next_steps)

        return CurriculumState(
            level=next_level,
            state=EpisodeLengthCurriculumState(
                step_counter=next_steps,
                ema_episode_length=ema,
            ),
        )

    def get_initial_state(self, rng: PRNGKeyArray) -> CurriculumState[EpisodeLengthCurriculumState]:
        return CurriculumState(
            level=jnp.array(self.min_level, dtype=jnp.float32),
            state=EpisodeLengthCurriculumState(
                step_counter=jnp.array(self.min_level_steps, dtype=jnp.int32),
                ema_episode_length=jnp.array(0.0, dtype=jnp.float32),
            ),
        )


@jax.tree_util.register_dataclass
@dataclass(frozen=True)
class DistanceFromOriginCurriculumState:
    step_counter: Array
    ema_distance: Array


@attrs.define(frozen=True, kw_only=True)
class DistanceFromOriginCurriculum(Curriculum[DistanceFromOriginCurriculumState]):
    """Curriculum that updates the distance from the origin based on thresholds."""

    num_levels: int = attrs.field(default=100, validator=attrs.validators.ge(1))
    increase_threshold: float = attrs.field(default=3.0, validator=attrs.validators.ge(0.0))
    decrease_threshold: float = attrs.field(default=3.0, validator=attrs.validators.ge(0.0))
    min_level_steps: int = attrs.field(default=1, validator=attrs.validators.ge(1))
    min_level: float = attrs.field(default=0.0, validator=attrs.validators.ge(0.0))
    ema_decay: float = attrs.field(default=0.9, validator=attrs.validators.ge(0.0))

    def __call__(
        self,
        trajectory: Trajectory,
        rewards: RewardState,
        training_state: xax.State,
        prev_state: CurriculumState[DistanceFromOriginCurriculumState],
    ) -> CurriculumState[DistanceFromOriginCurriculumState]:
        # Calculate current distance from origin
        distance = jnp.linalg.norm(trajectory.qpos[..., :3], axis=-1).max()
        ema = self.ema_decay * prev_state.state.ema_distance + (1 - self.ema_decay) * distance

        # Get current level and steps
        current_level = prev_state.level
        steps = jnp.maximum(prev_state.state.step_counter - 1, 0)

        # Update level based on thresholds
        new_level_if_enough_steps = jnp.where(
            ema > self.increase_threshold,
            jnp.minimum(current_level + 1.0 / self.num_levels, 1.0),
            jnp.where(
                ema < self.decrease_threshold,
                jnp.maximum(current_level - 1.0 / self.num_levels, 0.0),
                current_level,
            ),
        )

        new_level = jnp.where(steps == 0, new_level_if_enough_steps, current_level)

        # Reset steps if level changed
        steps = jnp.where(new_level != current_level, self.min_level_steps, steps)

        return CurriculumState(
            level=new_level,
            state=DistanceFromOriginCurriculumState(
                step_counter=steps,
                ema_distance=ema,
            ),
        )

    def get_initial_state(self, rng: PRNGKeyArray) -> CurriculumState[DistanceFromOriginCurriculumState]:
        return CurriculumState(
            level=jnp.array(self.min_level, dtype=jnp.float32),
            state=DistanceFromOriginCurriculumState(
                step_counter=jnp.array(self.min_level_steps, dtype=jnp.int32),
                ema_distance=jnp.array(0.0, dtype=jnp.float32),
            ),
        )


@attrs.define(frozen=True, kw_only=True)
class StepWhenSaturated(Curriculum[Array]):
    """Step the curriculum depending on how many deaths there are per episode.

    The logic here is that if there are fewer than `increase_threshold` deaths
    per episode on average, then we should increase the curriculum level.
    Similarly, if there are more than `decrease_threshold` deaths per episode
    on average, then we should decrease the curriculum level.

    We only allow the level to be changed every `min_level_steps` steps.
    """

    num_levels: int = attrs.field()
    increase_threshold: float = attrs.field()
    decrease_threshold: float = attrs.field()
    min_level_steps: int = attrs.field()
    min_level: float = attrs.field(default=0.0, validator=attrs.validators.ge(0.0))

    def __call__(
        self,
        trajectory: Trajectory,
        rewards: RewardState,
        training_state: xax.State,
        prev_state: CurriculumState[Array],
    ) -> CurriculumState[Array]:
        level, steps = prev_state.level, prev_state.state
        level, steps = jax.lax.cond(
            steps <= 0,
            lambda: self._step_level(level, trajectory),
            lambda: (level, steps - 1),
        )
        return CurriculumState(level=level, state=steps)

    def _step_level(self, level: Array, trajectory: Trajectory) -> tuple[Array, Array]:
        deaths = trajectory.done.sum(axis=-1).mean()
        should_increase = deaths < self.increase_threshold
        should_decrease = deaths > self.decrease_threshold
        delta = 1.0 / self.num_levels
        level = jnp.where(
            should_increase,
            level + delta,
            jnp.where(should_decrease, level - delta, level),
        )
        return jnp.clip(level, self.min_level, 1.0), jnp.array(self.min_level_steps, dtype=jnp.int32)

    def get_initial_state(self, rng: PRNGKeyArray) -> CurriculumState[Array]:
        return CurriculumState(
            level=jnp.array(self.min_level, dtype=jnp.float32),
            state=jnp.array(self.min_level_steps, dtype=jnp.int32),
        )



"""Base JAX centric environment class.

Designed to be easily extensible to any physics engine that separates
model and data.
"""

__all__ = [
    "PhysicsEngine",
    "MjxEngine",
    "MujocoEngine",
    "get_physics_engine",
    "engine_type_from_physics_model",
]

import logging
from abc import ABC, abstractmethod
from typing import Collection, Literal

import equinox as eqx
import jax
import jax.numpy as jnp
import mujoco
import xax
from jaxtyping import Array, PRNGKeyArray, PyTree
from mujoco import mjx

from ksim.actuators import Actuators, StatefulActuators
from ksim.debugging import JitLevel
from ksim.events import Event
from ksim.resets import Reset
from ksim.types import PhysicsModel, PhysicsState

logger = logging.getLogger(__name__)

EngineType = Literal["mjx", "mujoco"]


class PhysicsEngine(eqx.Module, ABC):
    """The role of an engine is simple: reset and step. Decoupled from data."""

    actuators: Actuators
    resets: Collection[Reset]
    events: Collection[Event]
    phys_steps_per_ctrl_steps: int
    min_action_latency_step: float
    max_action_latency_step: float
    drop_action_prob: float
    phys_steps_per_actuator_step: int

    def __init__(
        self,
        resets: Collection[Reset],
        events: Collection[Event],
        actuators: Actuators,
        phys_steps_per_ctrl_steps: int,
        min_action_latency_step: float,
        max_action_latency_step: float,
        drop_action_prob: float,
        phys_steps_per_actuator_step: int,
    ) -> None:
        """Initialize the MJX engine with resetting and actuators."""
        self.actuators = actuators
        self.resets = resets
        self.events = events
        self.phys_steps_per_ctrl_steps = phys_steps_per_ctrl_steps
        self.min_action_latency_step = min_action_latency_step
        self.max_action_latency_step = max_action_latency_step
        self.drop_action_prob = drop_action_prob
        self.phys_steps_per_actuator_step = phys_steps_per_actuator_step

    @abstractmethod
    def reset(self, physics_model: PhysicsModel, curriculum_level: Array, rng: PRNGKeyArray) -> PhysicsState:
        """Reset the engine and return the physics data."""

    @abstractmethod
    def step(
        self,
        action: Array,
        physics_model: PhysicsModel,
        physics_state: PhysicsState,
        curriculum_level: Array,
        rng: PRNGKeyArray,
    ) -> PhysicsState:
        """Step the engine and return the updated physics data."""

    def _reset_events(self, rng: PRNGKeyArray) -> xax.FrozenDict[str, PyTree]:
        event_states: dict[str, PyTree] = {}
        for event in self.events:
            rng, event_rng = jax.random.split(rng)
            event_states[event.event_name] = event.get_initial_event_state(event_rng)
        return xax.FrozenDict(event_states)


class MjxEngine(PhysicsEngine):
    """Defines an engine for MJX models."""

    @xax.jit(static_argnames=["self"])
    def reset(self, physics_model: mjx.Model, curriculum_level: Array, rng: PRNGKeyArray) -> PhysicsState:
        mjx_data = mjx.make_data(physics_model)

        for reset in self.resets:
            rng, reset_rng = jax.random.split(rng)
            mjx_data = reset(mjx_data, curriculum_level, reset_rng)

        mjx_data = mjx.forward(physics_model, mjx_data)
        assert isinstance(mjx_data, mjx.Data)

        # Zeros out some non-zeroed quantities.
        mjx_data = mjx_data.replace(
            qvel=jnp.zeros_like(mjx_data.qvel),
            qacc=jnp.zeros_like(mjx_data.qacc),
        )

        default_action = self.actuators.get_default_action(mjx_data)

        # Gets the initial actuator state for stateful actuators.
        actuator_state = (
            self.actuators.get_initial_state(mjx_data, rng) if isinstance(self.actuators, StatefulActuators) else None
        )

        rng, latency_rng = jax.random.split(rng)

        return PhysicsState(
            data=mjx_data,
            most_recent_action=default_action,
            event_states=self._reset_events(rng),
            actuator_state=actuator_state,
            action_latency=jax.random.uniform(
                latency_rng,
                minval=self.min_action_latency_step * curriculum_level,
                maxval=self.max_action_latency_step * curriculum_level,
            ),
        )

    @xax.jit(static_argnames=["self"], jit_level=JitLevel.MJX)
    def _physics_step(self, physics_model: mjx.Model, data: mjx.Data) -> mjx.Data:
        # Just performs the MJX step, but wraps it in it's own JIT which can be
        # cached to prevent heavy recompilation every time the rewards or
        # events change.
        return mjx.step(physics_model, data)

    @xax.jit(static_argnames=["self"], jit_level=JitLevel.ENGINE)
    def step(
        self,
        action: Array,
        physics_model: mjx.Model,
        physics_state: PhysicsState,
        curriculum_level: Array,
        rng: PRNGKeyArray,
    ) -> PhysicsState:
        mjx_data = physics_state.data
        phys_steps_per_ctrl_steps = self.phys_steps_per_ctrl_steps
        prev_action = physics_state.most_recent_action

        # Randomly drops some actions.
        rng, drop_rng = jax.random.split(rng)
        drop_action = jax.random.bernoulli(drop_rng, self.drop_action_prob * curriculum_level, shape=action.shape)
        action = jnp.where(drop_action, prev_action, action)

        def move_physics(
            carry: tuple[mjx.Data, Array, xax.FrozenDict[str, PyTree], PyTree, Array],
            rng: PRNGKeyArray,
        ) -> tuple[tuple[mjx.Data, Array, xax.FrozenDict[str, PyTree], PyTree, Array], None]:
            data, step_num, event_states, actuator_state, last_torques = carry

            # Randomly apply the action with some latency.
            prct = jnp.clip(step_num - physics_state.action_latency, 0.0, 1.0)
            ctrl = prev_action * (1.0 - prct) + action * prct

            should_update = jnp.mod(step_num, self.phys_steps_per_actuator_step) == 0

            def update_actuators(prev_actuator_state: PyTree) -> tuple[Array, PyTree]:
                if isinstance(self.actuators, StatefulActuators):
                    torques, actuator_state = self.actuators.get_stateful_ctrl(
                        action=ctrl,
                        physics_data=data,
                        actuator_state=prev_actuator_state,
                        rng=rng,
                    )
                else:
                    torques = self.actuators.get_ctrl(action=ctrl, physics_data=data, rng=rng)
                    actuator_state = prev_actuator_state
                return torques, actuator_state

            torques, actuator_state = jax.lax.cond(
                should_update,
                lambda: update_actuators(actuator_state),
                lambda: (last_torques, actuator_state),
            )

            data = data.replace(ctrl=torques)

            # Apply the events.
            new_event_states = {}
            for event in self.events:
                rng, event_rng = jax.random.split(rng)
                data, new_event_state = event(
                    model=physics_model,
                    data=data,
                    event_state=event_states[event.event_name],
                    curriculum_level=curriculum_level,
                    rng=event_rng,
                )
                new_event_states[event.event_name] = new_event_state

            new_data = self._physics_step(physics_model, data)
            return (new_data, step_num + 1.0, xax.FrozenDict(new_event_states), actuator_state, torques), None

        # Runs the model for N steps.
        (mjx_data, *_, event_info, actuator_state_final, _), _ = xax.scan(
            move_physics,
            # Note that we pass in 0 for initial last computed torques but step 0 will always update.
            (
                mjx_data,
                jnp.array(0.0),
                physics_state.event_states,
                physics_state.actuator_state,
                jnp.zeros_like(physics_state.data.ctrl),
            ),
            jax.random.split(rng, phys_steps_per_ctrl_steps),
            jit_level=JitLevel.ENGINE,
        )

        return PhysicsState(
            data=mjx_data,
            most_recent_action=action,
            event_states=xax.FrozenDict(event_info),
            actuator_state=actuator_state_final,
            action_latency=physics_state.action_latency,
        )


class MujocoEngine(PhysicsEngine):
    """Defines an engine for MuJoCo models."""

    def reset(self, physics_model: mujoco.MjModel, curriculum_level: Array, rng: PRNGKeyArray) -> PhysicsState:
        mj_data = mujoco.MjData(physics_model)

        for reset in self.resets:
            rng, reset_rng = jax.random.split(rng)
            mj_data = reset(mj_data, curriculum_level, reset_rng)

        mujoco.mj_forward(physics_model, mj_data)

        # Zeros out some non-zeroed quantities.
        mj_data.qvel[:] = 0.0
        mj_data.qacc[:] = 0.0

        default_action = self.actuators.get_default_action(mj_data)
        actuator_state = (
            self.actuators.get_initial_state(mj_data, rng) if isinstance(self.actuators, StatefulActuators) else None
        )

        rng, latency_rng = jax.random.split(rng)

        return PhysicsState(
            data=mj_data,
            most_recent_action=default_action,
            event_states=self._reset_events(rng),
            actuator_state=actuator_state,
            action_latency=jax.random.uniform(
                latency_rng,
                minval=self.min_action_latency_step * curriculum_level,
                maxval=self.max_action_latency_step * curriculum_level,
            ),
        )

    def step(
        self,
        action: Array,
        physics_model: mujoco.MjModel,
        physics_state: PhysicsState,
        curriculum_level: Array,
        rng: PRNGKeyArray,
    ) -> PhysicsState:
        data = physics_state.data

        if not isinstance(data, mujoco.MjData):
            raise ValueError("Mujoco data is not a MjData")

        mujoco.mj_forward(physics_model, data)
        phys_steps_per_ctrl_steps = self.phys_steps_per_ctrl_steps
        prev_action = physics_state.most_recent_action

        # Randomly drops some actions.
        rng, drop_rng = jax.random.split(rng)
        drop_action = jax.random.bernoulli(drop_rng, self.drop_action_prob * curriculum_level, shape=action.shape)
        action = jnp.where(drop_action, prev_action, action)

        event_states = physics_state.event_states
        actuator_state = physics_state.actuator_state

        last_torques = jnp.zeros_like(action)

        for step_num in range(phys_steps_per_ctrl_steps):
            # Randomly apply the action with some latency.
            prct = jnp.clip(step_num - physics_state.action_latency, 0.0, 1.0)
            ctrl = prev_action * (1.0 - prct) + action * prct

            if (step_num % self.phys_steps_per_actuator_step) == 0:
                if isinstance(self.actuators, StatefulActuators):
                    torques, actuator_state = self.actuators.get_stateful_ctrl(
                        action=ctrl,
                        physics_data=data,
                        actuator_state=actuator_state,
                        rng=rng,
                    )
                else:
                    torques = self.actuators.get_ctrl(action=ctrl, physics_data=data, rng=rng)
                last_torques = torques
            else:
                torques = last_torques

            data.ctrl[:] = torques

            # Apply the events.
            new_event_states = {}
            for event in self.events:
                data, new_event_state = event(
                    model=physics_model,
                    data=data,
                    event_state=event_states[event.event_name],
                    curriculum_level=curriculum_level,
                    rng=rng,
                )
                new_event_states[event.event_name] = new_event_state
            event_states = xax.FrozenDict(new_event_states)

            mujoco.mj_step(physics_model, data)

        return PhysicsState(
            data=data,
            most_recent_action=action,
            event_states=event_states,
            actuator_state=actuator_state,
            action_latency=physics_state.action_latency,
        )


def get_physics_engine(
    engine_type: EngineType,
    resets: Collection[Reset],
    events: Collection[Event],
    actuators: Actuators,
    *,
    dt: float,
    ctrl_dt: float,
    action_latency_range: tuple[float, float],
    drop_action_prob: float,
    actuator_update_dt: float | None,
) -> PhysicsEngine:
    min_action_latency, max_action_latency = action_latency_range
    if min_action_latency < 0:
        raise ValueError("`min_action_latency` must be non-negative")
    if min_action_latency > max_action_latency:
        raise ValueError("`min_action_latency` must be less than or equal to `max_action_latency`")
    if max_action_latency > ctrl_dt:
        logger.warning("`max_action_latency=%f` is greater than `ctrl_dt=%f`", max_action_latency, ctrl_dt)
    if (remainder := (ctrl_dt - round(ctrl_dt / dt) * dt)) > 1e-6:
        logger.warning("`ctrl_dt=%f` is not a multiple of `dt=%f` (remainder=%f)", ctrl_dt, dt, remainder)
    if drop_action_prob < 0 or drop_action_prob >= 1:
        raise ValueError("`drop_action_prob` must be between 0 and 1")

    # Converts to steps.
    min_action_latency_step = min_action_latency / dt
    max_action_latency_step = max_action_latency / dt
    phys_steps_per_ctrl_steps = round(ctrl_dt / dt)

    if actuator_update_dt is not None:
        phys_steps_per_actuator_step = max(1, round(actuator_update_dt / dt))
    else:
        phys_steps_per_actuator_step = 1

    match engine_type:
        case "mujoco":
            return MujocoEngine(
                resets=resets,
                events=events,
                actuators=actuators,
                min_action_latency_step=min_action_latency_step,
                max_action_latency_step=max_action_latency_step,
                phys_steps_per_ctrl_steps=phys_steps_per_ctrl_steps,
                drop_action_prob=drop_action_prob,
                phys_steps_per_actuator_step=phys_steps_per_actuator_step,
            )

        case "mjx":
            return MjxEngine(
                resets=resets,
                events=events,
                actuators=actuators,
                min_action_latency_step=min_action_latency_step,
                max_action_latency_step=max_action_latency_step,
                phys_steps_per_ctrl_steps=phys_steps_per_ctrl_steps,
                drop_action_prob=drop_action_prob,
                phys_steps_per_actuator_step=phys_steps_per_actuator_step,
            )

        case _:
            raise ValueError(f"Unsupported physics model type: {engine_type}")


def engine_type_from_physics_model(physics_model: PhysicsModel) -> EngineType:
    if isinstance(physics_model, mujoco.MjModel):
        return "mujoco"
    if isinstance(physics_model, mjx.Model):
        return "mjx"
    raise ValueError(f"Unsupported physics model type: {type(physics_model)}")



"""Tests for rewards in the ksim package."""

import jax.numpy as jnp
import xax

import ksim


def create_test_trajectory(
    actions: jnp.ndarray,
    done: jnp.ndarray | None = None,
    timesteps: int | None = None,
    action_dims: int = 3,
) -> ksim.Trajectory:
    """Create a test trajectory with given actions.

    Args:
        actions: Action array of shape (timesteps, action_dims)
        done: Optional done mask of shape (timesteps,). If None, all False.
        timesteps: Number of timesteps. If None, inferred from actions.
        action_dims: Number of action dimensions.

    Returns:
        A ksim.Trajectory object for testing.
    """
    if timesteps is None:
        timesteps = actions.shape[0]

    if done is None:
        done = jnp.zeros(timesteps, dtype=bool)

    dummy_qpos = jnp.zeros((timesteps, 10))
    dummy_qvel = jnp.zeros((timesteps, 9))
    dummy_xpos = jnp.zeros((timesteps, 5, 3))
    dummy_xquat = jnp.zeros((timesteps, 5, 4))
    dummy_ctrl = jnp.zeros((timesteps, action_dims))
    dummy_timestep = jnp.arange(timesteps, dtype=jnp.float32)
    dummy_success = jnp.zeros(timesteps, dtype=bool)

    return ksim.Trajectory(
        qpos=dummy_qpos,
        qvel=dummy_qvel,
        xpos=dummy_xpos,
        xquat=dummy_xquat,
        ctrl=dummy_ctrl,
        obs=xax.FrozenDict({}),
        command=xax.FrozenDict({}),
        event_state=xax.FrozenDict({}),
        action=actions,
        done=done,
        success=dummy_success,
        timestep=dummy_timestep,
        termination_components=xax.FrozenDict({}),
        aux_outputs=xax.FrozenDict({}),
    )


class TestActionVelocityPenalty:
    """Test suite for ActionVelocityPenalty."""

    def test_constant_action_zero_penalty(self) -> None:
        """Test that constant actions result in zero velocity penalty."""
        penalty = ksim.ActionVelocityPenalty(scale=-1.0)

        constant_actions = jnp.ones((5, 3)) * 2.0
        trajectory = create_test_trajectory(constant_actions)

        reward_per_timestep = penalty.get_reward(trajectory)

        expected_per_timestep = jnp.zeros(5)
        assert jnp.allclose(reward_per_timestep, expected_per_timestep, atol=1e-6)

    def test_linear_action_change_constant_velocity(self) -> None:
        """Test linear action changes result in constant velocity penalty."""
        penalty = ksim.ActionVelocityPenalty(scale=-1.0, norm="l2")

        # Linear increase: actions go from [0,0,0] to [4,4,4] over 5 steps
        actions = jnp.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 1.0, 1.0],
                [2.0, 2.0, 2.0],
                [3.0, 3.0, 3.0],
                [4.0, 4.0, 4.0],
            ]
        )
        trajectory = create_test_trajectory(actions)

        reward_per_timestep = penalty.get_reward(trajectory)

        # With masking, first timestep is 0, rest have constant velocity penalty
        # Velocities from timestep 1 onwards: [1,1,1], [1,1,1], [1,1,1], [1,1,1]
        # L2 element-wise squares: [1,1,1], [1,1,1], [1,1,1], [1,1,1]
        # Mean over dimensions: 1, 1, 1, 1
        expected_per_timestep = jnp.array([0.0, 1.0, 1.0, 1.0, 1.0])
        assert jnp.allclose(reward_per_timestep, expected_per_timestep, atol=1e-6)

    def test_step_action_change(self) -> None:
        """Test step change in actions."""
        penalty = ksim.ActionVelocityPenalty(scale=-1.0, norm="l1")

        # Step change: constant then sudden jump
        actions = jnp.array(
            [
                [1.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [3.0, 2.0, -1.0],
                [3.0, 2.0, -1.0],
            ]
        )
        trajectory = create_test_trajectory(actions)

        reward_per_timestep = penalty.get_reward(trajectory)

        # With masking, first timestep is 0
        # Velocity changes from timestep 1: [0,0,0], [0,0,0], [2,2,-1], [0,0,0]
        # L1 element-wise absolute values: [0,0,0], [0,0,0], [2,2,1], [0,0,0]
        # Mean over dimensions: 0, 0, 5/3, 0
        expected_per_timestep = jnp.array([0.0, 0.0, 0.0, 5.0 / 3.0, 0.0])
        assert jnp.allclose(reward_per_timestep, expected_per_timestep, atol=1e-6)

    def test_done_mask_resets_velocity(self) -> None:
        """Test that done episodes reset action velocity calculations."""
        penalty = ksim.ActionVelocityPenalty(scale=-1.0, norm="l2")

        actions = jnp.array(
            [
                [0.0, 0.0, 0.0],
                [2.0, 0.0, 0.0],
                [4.0, 0.0, 0.0],
                [5.0, 0.0, 0.0],
            ]
        )
        done = jnp.array([False, True, False, False])
        trajectory = create_test_trajectory(actions, done=done)

        reward_per_timestep = penalty.get_reward(trajectory)

        # With masking, first timestep is 0
        # Velocities from timestep 1: [2,0,0], [0,0,0] (reset), [1,0,0]
        # L2 element-wise squares: [4,0,0], [0,0,0], [1,0,0]
        # Mean over dimensions: 4/3, 0, 1/3
        expected_per_timestep = jnp.array([0.0, 4.0 / 3.0, 0.0, 1.0 / 3.0])
        assert jnp.allclose(reward_per_timestep, expected_per_timestep, atol=1e-6)


class TestActionAccelerationPenalty:
    """Test suite for ActionAccelerationPenalty."""

    def test_constant_velocity_zero_acceleration(self) -> None:
        """Test that constant velocity (linear actions) gives zero acceleration penalty."""
        penalty = ksim.ActionAccelerationPenalty(scale=-1.0)

        # Linear actions: constant velocity
        actions = jnp.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 1.0, 1.0],
                [2.0, 2.0, 2.0],
                [3.0, 3.0, 3.0],
                [4.0, 4.0, 4.0],
            ]
        )
        trajectory = create_test_trajectory(actions)

        reward_per_timestep = penalty.get_reward(trajectory)

        # With masking, first two timesteps are 0
        # True constant velocity should give zero acceleration
        expected_per_timestep = jnp.array([0.0, 0.0, 0.0, 0.0, 0.0])
        assert jnp.allclose(reward_per_timestep, expected_per_timestep, atol=1e-6)

    def test_quadratic_action_constant_acceleration(self) -> None:
        """Test quadratic action progression gives constant acceleration."""
        penalty = ksim.ActionAccelerationPenalty(scale=-1.0, norm="l2")

        # Quadratic progression: 0, 1, 4, 9, 16 (squared values)
        actions = jnp.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 1.0, 1.0],
                [4.0, 4.0, 4.0],
                [9.0, 9.0, 9.0],
                [16.0, 16.0, 16.0],
            ]
        )
        trajectory = create_test_trajectory(actions)

        reward_per_timestep = penalty.get_reward(trajectory)

        # With masking, first two timesteps are 0
        # True quadratic progression: velocities [1,3,5,7], accelerations [2,2,2]
        # L2 element-wise squares: [4,4,4] for timesteps 2,3,4
        # Mean over dimensions: 4, 4, 4
        expected_per_timestep = jnp.array([0.0, 0.0, 4.0, 4.0, 4.0])
        assert jnp.allclose(reward_per_timestep, expected_per_timestep, atol=1e-6)

    def test_acceleration_step_change(self) -> None:
        """Test step change in acceleration."""
        penalty = ksim.ActionAccelerationPenalty(scale=-1.0, norm="l1")

        # Velocities: constant, then linear increase
        actions = jnp.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],  # vel = [1,0,0]
                [2.0, 0.0, 0.0],  # vel = [1,0,0]
                [2.0, 0.0, 0.0],  # vel = [0,0,0]
                [1.0, 0.0, 0.0],  # vel = [-1,0,0]
            ]
        )
        trajectory = create_test_trajectory(actions)

        reward_per_timestep = penalty.get_reward(trajectory)

        # With masking, first two timesteps are 0
        # True velocities: [1,1,0,-1], accelerations: [0,-1,-1]
        # L1 element-wise absolute values: [0,1,1] for timesteps 2,3,4
        # Mean over dimensions: 0, 1/3, 1/3
        expected_per_timestep = jnp.array([0.0, 0.0, 0.0, 1.0 / 3.0, 1.0 / 3.0])
        assert jnp.allclose(reward_per_timestep, expected_per_timestep, atol=1e-6)


class TestActionJerkPenalty:
    """Test suite for ActionJerkPenalty."""

    def test_constant_acceleration_zero_jerk(self) -> None:
        """Test that constant acceleration gives zero jerk penalty."""
        penalty = ksim.ActionJerkPenalty(scale=-1.0)

        # Quadratic actions: constant acceleration
        actions = jnp.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 1.0, 1.0],
                [4.0, 4.0, 4.0],
                [9.0, 9.0, 9.0],
                [16.0, 16.0, 16.0],
                [25.0, 25.0, 25.0],
            ]
        )
        trajectory = create_test_trajectory(actions)

        reward_per_timestep = penalty.get_reward(trajectory)

        # With masking, first three timesteps are 0
        # True quadratic progression should give zero jerk
        expected_per_timestep = jnp.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        assert jnp.allclose(reward_per_timestep, expected_per_timestep, atol=1e-6)

    def test_cubic_action_constant_jerk(self) -> None:
        """Test cubic action progression gives constant jerk."""
        penalty = ksim.ActionJerkPenalty(scale=-1.0, norm="l2")

        # Cubic progression: t^3 for t = 0,1,2,3,4,5
        t = jnp.arange(6)
        cubic_values = t**3
        actions = jnp.stack([cubic_values, cubic_values, cubic_values], axis=1)
        trajectory = create_test_trajectory(actions)

        reward_per_timestep = penalty.get_reward(trajectory)

        # With masking, first three timesteps are 0
        # True cubic progression: velocities [1,7,19,37,61], accelerations [6,12,18,24], jerks [6,6,6]
        # L2 element-wise squares: [36,36,36] for timesteps 3,4,5
        # Mean over dimensions: 36, 36, 36
        expected_per_timestep = jnp.array([0.0, 0.0, 0.0, 36.0, 36.0, 36.0])
        assert jnp.allclose(reward_per_timestep, expected_per_timestep, atol=1e-5)



"""Defines debugging utility functions."""

__all__ = [
    "JitLevel",
]


class JitLevel:
    OUTER_LOOP = 10
    RL_CORE = 20
    UNROLL = 30
    INITIALIZATION = 40
    ENGINE = 50
    HELPER_FUNCTIONS = 60
    MJX = 70



"""Defines a base interface for defining reward functions."""

__all__ = [
    "MonotonicFn",
    "norm_to_reward",
    "Reward",
    "StatefulReward",
    "StayAliveReward",
    "LinearVelocityReward",
    "LinearVelocityPenalty",
    "NaiveForwardReward",
    "NaiveForwardOrientationReward",
    "AngularVelocityReward",
    "AngularVelocityPenalty",
    "XYAngularVelocityPenalty",
    "BaseHeightReward",
    "BaseHeightRangeReward",
    "ActionVelocityPenalty",
    "ActionAccelerationPenalty",
    "ActionJerkPenalty",
    "JointVelocityPenalty",
    "JointAccelerationPenalty",
    "JointJerkPenalty",
    "AvoidLimitsPenalty",
    "CtrlPenalty",
    "JointDeviationPenalty",
    "FlatBodyReward",
    "PositionTrackingReward",
    "UprightReward",
    "LinkAccelerationPenalty",
    "LinkJerkPenalty",
    "JoystickReward",
    "ReachabilityPenalty",
]

import functools
import logging
from abc import ABC, abstractmethod
from typing import Collection, Literal, Self

import attrs
import chex
import jax.numpy as jnp
import xax
from jaxtyping import Array, PRNGKeyArray, PyTree

from ksim.types import PhysicsModel, Trajectory
from ksim.utils.mujoco import get_body_data_idx_from_name, get_qpos_data_idxs_by_name
from ksim.utils.validators import (
    CartesianIndex,
    cartesian_index_to_dim,
    dimension_index_tuple_validator,
    norm_validator,
)
from ksim.vis import Marker

logger = logging.getLogger(__name__)

MonotonicFn = Literal["exp", "inv", "sigmoid"]


def norm_to_reward(value: Array, temp: float = 1.0, monotonic_fn: MonotonicFn = "inv") -> Array:
    """Helper function for converting from a norm to a reward.

    Args:
        value: The value (usually a norm) to convert to a reward.
        temp: The temperature to use for the conversion. Higher temperatures
            will make the reward drop off less steeply.
        monotonic_fn: The monotonic function to use for the conversion.

    Returns:
        The reward.
    """
    match monotonic_fn:
        case "inv":
            return 1.0 / (value / temp + 1.0)
        case "exp":
            return jnp.exp(-value / temp)
        case "sigmoid":
            return 1.0 / (1.0 + jnp.exp(-value / temp))
        case _:
            raise ValueError(f"Invalid monotonic function: {monotonic_fn}")


def reward_scale_validator(inst: "Reward", attr: attrs.Attribute, value: float) -> None:
    # Reward function classes should end with either "Reward" or "Penalty",
    # which we use here to check if the scale is positive or negative.
    if inst.reward_name.lower().endswith("reward"):
        if value < 0:
            raise RuntimeError(f"Reward function {inst.reward_name} has a negative scale {value}")
    elif inst.reward_name.lower().endswith("penalty"):
        if value > 0:
            raise RuntimeError(f"Penalty function {inst.reward_name} has a positive scale {value}")
    else:
        logger.warning("Reward function %s does not end with 'Reward' or 'Penalty': %f", inst.reward_name, value)


def index_to_dims(index: CartesianIndex | tuple[CartesianIndex, ...]) -> tuple[int, ...]:
    indices = index if isinstance(index, tuple) else (index,)
    return tuple(cartesian_index_to_dim(index) for index in indices)


@attrs.define(frozen=True, kw_only=True)
class Reward(ABC):
    """Base class for defining reward functions."""

    scale: float = attrs.field(validator=reward_scale_validator)
    scale_by_curriculum: bool = attrs.field(default=False)

    @abstractmethod
    def get_reward(self, trajectory: Trajectory) -> Array:
        """Get the reward for a single trajectory.

        Args:
            trajectory: The trajectory to get the reward for.

        Returns:
            An array of shape (time) containing the reward for each timestep.
        """

    def get_markers(self) -> Collection[Marker]:
        """Get the markers for the reward, optionally overridable."""
        return []

    def get_name(self) -> str:
        return xax.camelcase_to_snakecase(self.__class__.__name__)

    @functools.cached_property
    def reward_name(self) -> str:
        return self.get_name()


@attrs.define(frozen=True, kw_only=True)
class StatefulReward(Reward):
    """Reward that requires state from the previous timestep."""

    def get_reward(self, trajectory: Trajectory) -> Array:
        raise NotImplementedError("StatefulReward should use `get_reward_stateful` instead.")

    @abstractmethod
    def initial_carry(self, rng: PRNGKeyArray) -> PyTree:
        """Initial reward carry for the trajectory, optionally overridable.

        Some rewards require information from the same episode in a previous
        rollout. E.g. a reward could require the last time the robot was in
        contact with the ground. This function simply returns the initial reward
        carry, which is `None` by default.
        """

    @abstractmethod
    def get_reward_stateful(self, trajectory: Trajectory, reward_carry: PyTree) -> tuple[Array, PyTree]:
        """Get the reward for a single trajectory.

        This is the same as `get_reward`, but it also takes in the reward carry
        from the previous timestep.

        Args:
            trajectory: The trajectory to get the reward for.
            reward_carry: The reward carry from the previous timestep.

        Returns:
            A tuple containing the reward for the current timestep and the
            reward carry for the next timestep.
        """


@attrs.define(frozen=True, kw_only=True)
class StayAliveReward(Reward):
    """Reward for staying alive.

    This provides a reward for staying alive, with a negative penalty on
    termination. These values are balanced by the balance parameter - a larger
    value will increase the relative penalty for termination.
    """

    balance: float = attrs.field(default=10.0)
    success_reward: float | None = attrs.field(default=None)

    def get_reward(self, trajectory: Trajectory) -> Array:
        reward = jnp.where(
            trajectory.done,
            jnp.where(
                trajectory.success,
                1.0 / self.balance if self.success_reward is None else self.success_reward,
                -1.0,
            ),
            1.0 / self.balance,
        )
        return reward


@attrs.define(frozen=True, kw_only=True)
class LinearVelocityReward(Reward):
    """Penalty for how fast the robot is moving in the z-direction."""

    index: CartesianIndex | tuple[CartesianIndex, ...] = attrs.field(validator=dimension_index_tuple_validator)
    clip_min: float | None = attrs.field(default=None)
    clip_max: float | None = attrs.field(default=None)
    norm: xax.NormType = attrs.field(default="l2", validator=norm_validator)
    in_robot_frame: bool = attrs.field(default=True)

    def get_reward(self, trajectory: Trajectory) -> Array:
        dims = index_to_dims(self.index)
        linvel = trajectory.qvel[..., :3]
        if self.in_robot_frame:
            # Same as reading from a velocimeter attached to base.
            linvel = xax.rotate_vector_by_quat(linvel, trajectory.qpos[..., 3:7], inverse=True)
        dimvel = linvel[..., dims].clip(self.clip_min, self.clip_max).mean(axis=-1)
        return xax.get_norm(dimvel, self.norm)

    def get_name(self) -> str:
        indices = self.index if isinstance(self.index, tuple) else (self.index,)
        return f"{''.join(indices)}_{super().get_name()}"


@attrs.define(frozen=True, kw_only=True)
class LinearVelocityPenalty(LinearVelocityReward): ...


@attrs.define(frozen=True, kw_only=True)
class NaiveForwardReward(Reward):
    """Simple reward for moving forward in the X-direction."""

    clip_min: float | None = attrs.field(default=None)
    clip_max: float | None = attrs.field(default=None)
    in_robot_frame: bool = attrs.field(default=True)

    def get_reward(self, trajectory: Trajectory) -> Array:
        linvel = trajectory.qvel[..., :3]
        if self.in_robot_frame:
            # Same as reading from a velocimeter attached to base.
            linvel = xax.rotate_vector_by_quat(linvel, trajectory.qpos[..., 3:7], inverse=True)
        dimvel = linvel[..., 0].clip(self.clip_min, self.clip_max)
        return dimvel


@attrs.define(frozen=True, kw_only=True)
class NaiveForwardOrientationReward(NaiveForwardReward):
    """Simple reward for keeping the robot oriented in the X-direction."""

    def get_reward(self, trajectory: Trajectory) -> Array:
        quat = trajectory.qpos[..., 3:7]
        forward_vec = jnp.array([1.0, 0.0, 0.0])
        forward_vec = xax.rotate_vector_by_quat(forward_vec, quat, inverse=True)
        return forward_vec[..., 0] - jnp.linalg.norm(forward_vec[..., 1:], axis=-1)


@attrs.define(frozen=True, kw_only=True)
class AngularVelocityReward(Reward):
    """Penalty for how fast the robot is rotating in the xy-plane."""

    index: CartesianIndex | tuple[CartesianIndex, ...] = attrs.field(validator=dimension_index_tuple_validator)
    clip_min: float | None = attrs.field(default=None)
    clip_max: float | None = attrs.field(default=None)
    norm: xax.NormType = attrs.field(default="l2", validator=norm_validator)
    in_robot_frame: bool = attrs.field(default=True)

    def get_reward(self, trajectory: Trajectory) -> Array:
        dims = index_to_dims(self.index)
        angvel = trajectory.qvel[..., 3:6]
        if self.in_robot_frame:
            angvel = xax.rotate_vector_by_quat(angvel, trajectory.qpos[..., 3:7], inverse=True)
        dimvel = angvel[..., dims].clip(self.clip_min, self.clip_max).mean(axis=-1)
        return xax.get_norm(dimvel, self.norm)

    def get_name(self) -> str:
        indices = self.index if isinstance(self.index, tuple) else (self.index,)
        return f"{''.join(indices)}_{super().get_name()}"


@attrs.define(frozen=True, kw_only=True)
class AngularVelocityPenalty(AngularVelocityReward): ...


@attrs.define(frozen=True, kw_only=True)
class XYAngularVelocityPenalty(AngularVelocityReward):
    index: CartesianIndex | tuple[CartesianIndex, ...] = attrs.field(
        default=("x", "y"),
        validator=dimension_index_tuple_validator,
    )


@attrs.define(frozen=True, kw_only=True)
class BaseHeightReward(Reward):
    """Penalty for deviating from the base height target."""

    height_target: float = attrs.field()
    norm: xax.NormType = attrs.field(default="l1", validator=norm_validator)
    temp: float = attrs.field(default=1.0)
    monotonic_fn: MonotonicFn = attrs.field(default="inv")

    def get_reward(self, trajectory: Trajectory) -> Array:
        base_height = trajectory.qpos[..., 2]
        reward = norm_to_reward(xax.get_norm(base_height - self.height_target, self.norm), self.temp, self.monotonic_fn)
        return reward


@attrs.define(frozen=True, kw_only=True)
class BaseHeightRangeReward(Reward):
    """Incentivizes keeping the base height within a certain range."""

    z_lower: float = attrs.field()
    z_upper: float = attrs.field()
    dropoff: float = attrs.field()

    def get_reward(self, trajectory: Trajectory) -> Array:
        base_height = trajectory.qpos[..., 2]
        too_low = self.z_lower - base_height
        too_high = base_height - self.z_upper
        reward = (1.0 - jnp.maximum(too_low, too_high).clip(min=0.0) * self.dropoff).clip(min=0.0)
        return reward


@attrs.define(frozen=True, kw_only=True)
class ActionVelocityPenalty(Reward):
    """Penalty for first derivative change in consecutive actions."""

    norm: xax.NormType = attrs.field(default="l2", validator=norm_validator)

    def get_reward(self, trajectory: Trajectory) -> Array:
        actions = trajectory.action
        actions_zp = jnp.pad(actions, ((1, 0), (0, 0)), mode="edge")
        done = jnp.pad(trajectory.done, ((1, 0),), mode="edge")[..., :-1, None]
        actions_vel = jnp.where(done, 0.0, actions_zp[..., 1:, :] - actions_zp[..., :-1, :])
        penalty = xax.get_norm(actions_vel, self.norm).mean(axis=-1)

        mask = jnp.arange(len(penalty)) >= 1
        penalty = jnp.where(mask, penalty, 0.0)
        return penalty


@attrs.define(frozen=True, kw_only=True)
class ActionAccelerationPenalty(Reward):
    """Penalty for second derivative change in consecutive actions."""

    norm: xax.NormType = attrs.field(default="l2", validator=norm_validator)

    def get_reward(self, trajectory: Trajectory) -> Array:
        actions = trajectory.action
        actions_zp = jnp.pad(actions, ((2, 0), (0, 0)), mode="edge")
        done = jnp.pad(trajectory.done, ((2, 0),), mode="edge")[..., :-1, None]
        actions_vel = jnp.where(done, 0.0, actions_zp[..., 1:, :] - actions_zp[..., :-1, :])
        actions_acc = jnp.where(done[..., 1:, :], 0.0, actions_vel[..., 1:, :] - actions_vel[..., :-1, :])
        penalty = xax.get_norm(actions_acc, self.norm).mean(axis=-1)

        # Mask out timesteps affected by padding to prevent artificial artifacts
        # (e.g. jump from constant velocity to non-zero acceleration)
        mask = jnp.arange(len(penalty)) >= 2
        penalty = jnp.where(mask, penalty, 0.0)
        return penalty


@attrs.define(frozen=True, kw_only=True)
class ActionJerkPenalty(Reward):
    """Penalty for third derivative change in consecutive actions."""

    norm: xax.NormType = attrs.field(default="l2", validator=norm_validator)

    def get_reward(self, trajectory: Trajectory) -> Array:
        actions = trajectory.action
        actions_zp = jnp.pad(actions, ((3, 0), (0, 0)), mode="edge")
        done = jnp.pad(trajectory.done, ((3, 0),), mode="edge")[..., :-1, None]
        actions_vel = jnp.where(done, 0.0, actions_zp[..., 1:, :] - actions_zp[..., :-1, :])
        actions_acc = jnp.where(done[..., 1:, :], 0.0, actions_vel[..., 1:, :] - actions_vel[..., :-1, :])
        actions_jerk = jnp.where(done[..., 2:, :], 0.0, actions_acc[..., 1:, :] - actions_acc[..., :-1, :])
        penalty = xax.get_norm(actions_jerk, self.norm).mean(axis=-1)

        mask = jnp.arange(len(penalty)) >= 3
        penalty = jnp.where(mask, penalty, 0.0)
        return penalty


@attrs.define(frozen=True, kw_only=True)
class JointVelocityPenalty(Reward):
    """Penalty for how fast the joint angular velocities are changing."""

    norm: xax.NormType = attrs.field(default="l2", validator=norm_validator)

    def get_reward(self, trajectory: Trajectory) -> Array:
        done = trajectory.done[..., None]
        joint_vel = jnp.where(done, 0.0, trajectory.qvel[..., 6:])
        return xax.get_norm(joint_vel, self.norm).mean(axis=-1)


@attrs.define(frozen=True, kw_only=True)
class JointAccelerationPenalty(Reward):
    """Penalty for high joint accelerations."""

    norm: xax.NormType = attrs.field(default="l2", validator=norm_validator)

    def get_reward(self, trajectory: Trajectory) -> Array:
        qpos = trajectory.qpos[..., 7:]
        qpos_zp = jnp.pad(qpos, ((2, 0), (0, 0)), mode="edge")
        done = jnp.pad(trajectory.done, ((2, 0),), mode="edge")[..., :-1, None]
        qvel = jnp.where(done, 0.0, qpos_zp[..., 1:, :] - qpos_zp[..., :-1, :])
        qacc = jnp.where(done[..., 1:, :], 0.0, qvel[..., 1:, :] - qvel[..., :-1, :])
        penalty = xax.get_norm(qacc, self.norm).mean(axis=-1)
        return penalty


@attrs.define(frozen=True, kw_only=True)
class JointJerkPenalty(Reward):
    """Penalty for high joint jerks."""

    norm: xax.NormType = attrs.field(default="l2", validator=norm_validator)

    def get_reward(self, trajectory: Trajectory) -> Array:
        qpos = trajectory.qpos[..., 7:]
        qpos_zp = jnp.pad(qpos, ((3, 0), (0, 0)), mode="edge")
        done = jnp.pad(trajectory.done, ((3, 0),), mode="edge")[..., :-1, None]
        qvel = jnp.where(done, 0.0, qpos_zp[..., 1:, :] - qpos_zp[..., :-1, :])
        qacc = jnp.where(done[..., 1:, :], 0.0, qvel[..., 1:, :] - qvel[..., :-1, :])
        qjerk = jnp.where(done[..., 2:, :], 0.0, qacc[..., 1:, :] - qacc[..., :-1, :])
        penalty = xax.get_norm(qjerk, self.norm).mean(axis=-1)
        return penalty


def joint_limits_validator(inst: "AvoidLimitsPenalty", attr: attrs.Attribute, value: xax.HashableArray) -> None:
    arr = value.array
    if arr.ndim != 2 or arr.shape[1] != 2:
        raise ValueError(f"Joint range must have shape (n_joints, 2), got {arr.shape}")
    if not jnp.all(arr[..., 0] <= arr[..., 1]):
        raise ValueError(f"Joint range must be sorted, got {arr}")
    if not arr.dtype == jnp.float32:
        raise ValueError(f"Joint range must be a float array, got {arr.dtype}")


@attrs.define(frozen=True, kw_only=True)
class AvoidLimitsPenalty(Reward):
    """Reward for keeping the joint positions away from the joint limits."""

    joint_limits: xax.HashableArray = attrs.field(validator=joint_limits_validator)

    def get_reward(self, trajectory: Trajectory) -> Array:
        joint_pos = trajectory.qpos[..., 7:]
        joint_limits = self.joint_limits.array
        penalty = -jnp.clip(joint_pos - joint_limits[..., 0], None, 0.0)
        penalty += jnp.clip(joint_pos - joint_limits[..., 1], 0.0, None)
        return penalty.mean(axis=-1)

    @classmethod
    def create(
        cls,
        model: PhysicsModel,
        factor: float = 0.05,
        scale: float = 1.0,
        scale_by_curriculum: bool = False,
    ) -> Self:
        jnt_range = jnp.array(model.jnt_range)
        jnt_limited = jnp.array(model.jnt_limited, dtype=jnp.bool_)
        jnt_min, jnt_max = jnt_range[..., 0], jnt_range[..., 1]
        jnt_diff = (jnt_max - jnt_min) * factor
        jnt_min = jnt_min - jnt_diff
        jnt_max = jnt_max + jnt_diff
        jnt_min = jnp.where(jnt_limited, jnt_min, -jnp.inf)
        jnt_max = jnp.where(jnt_limited, jnt_max, jnp.inf)
        joint_limits = jnp.stack([jnt_min, jnt_max], axis=-1)
        return cls(
            joint_limits=xax.hashable_array(joint_limits[..., 1:, :]),
            scale=scale,
            scale_by_curriculum=scale_by_curriculum,
        )


@attrs.define(frozen=True, kw_only=True)
class CtrlPenalty(Reward):
    """Penalty for large torque commands."""

    norm: xax.NormType = attrs.field(default="l2")
    scales: tuple[float, ...] | None = attrs.field(default=None)

    def get_reward(self, trajectory: Trajectory) -> Array:
        ctrl = trajectory.ctrl
        if self.scales is not None:
            ctrl = ctrl / jnp.array(self.scales)
        return xax.get_norm(ctrl, self.norm).mean(axis=-1)

    @classmethod
    def create(cls, model: PhysicsModel, scale: float = -1.0, scale_by_curriculum: bool = False) -> Self:
        ctrl_min = model.actuator_ctrlrange[..., 0]
        ctrl_max = model.actuator_ctrlrange[..., 1]
        ctrl_range = (ctrl_max - ctrl_min) / 2.0
        ctrl_range_list = ctrl_range.flatten().tolist()
        return cls(
            scales=tuple(ctrl_range_list),
            scale=scale,
            scale_by_curriculum=scale_by_curriculum,
        )


@attrs.define(frozen=True, kw_only=True)
class JointDeviationPenalty(Reward):
    """Penalty for joint deviations from target positions."""

    norm: xax.NormType = attrs.field(default="l2")
    joint_indices: tuple[int, ...] = attrs.field()
    joint_targets: tuple[float, ...] = attrs.field()
    joint_weights: tuple[float, ...] | None = attrs.field(default=None)

    def get_reward(self, trajectory: Trajectory) -> Array:
        qpos_sel = trajectory.qpos[..., jnp.array(self.joint_indices) + 7]
        target = jnp.asarray(self.joint_targets)
        diff = qpos_sel - target

        if self.joint_weights is not None:
            diff *= jnp.array(self.joint_weights)

        penalty = xax.get_norm(diff, self.norm).sum(axis=-1)
        return penalty

    @classmethod
    def create(
        cls,
        physics_model: PhysicsModel,
        joint_names: tuple[str, ...],
        joint_targets: tuple[float, ...],
        scale: float = -1.0,
        joint_weights: tuple[float, ...] | None = None,
        scale_by_curriculum: bool = False,
    ) -> Self:
        joint_to_idx = get_qpos_data_idxs_by_name(physics_model)
        joint_indices = tuple([int(joint_to_idx[name][0]) - 7 for name in joint_names])
        return cls(
            joint_indices=joint_indices,
            joint_targets=joint_targets,
            joint_weights=joint_weights,
            scale=scale,
            scale_by_curriculum=scale_by_curriculum,
        )


@attrs.define(frozen=True, kw_only=True)
class FlatBodyReward(Reward):
    """Reward for keeping the body parallel to the ground."""

    body_indices: tuple[int, ...] = attrs.field()
    plane: tuple[float, float, float] = attrs.field(default=(0.0, 0.0, 1.0))
    norm: xax.NormType = attrs.field(default="l2", validator=norm_validator)

    def get_reward(self, trajectory: Trajectory) -> Array:
        body_quat = trajectory.xquat[..., self.body_indices, :]
        plane_vec = jnp.array(self.plane, dtype=body_quat.dtype)
        unit_vec = xax.rotate_vector_by_quat(plane_vec, body_quat, inverse=True)
        return jnp.einsum("...i,...i->...", unit_vec, plane_vec).mean(axis=-1)

    @classmethod
    def create(
        cls,
        physics_model: PhysicsModel,
        body_names: tuple[str, ...],
        plane: tuple[float, float, float] = (0.0, 0.0, 1.0),
        norm: xax.NormType = "l2",
        scale: float = 1.0,
        scale_by_curriculum: bool = False,
    ) -> Self:
        return cls(
            body_indices=tuple([get_body_data_idx_from_name(physics_model, name) for name in body_names]),
            plane=plane,
            norm=norm,
            scale=scale,
            scale_by_curriculum=scale_by_curriculum,
        )


@attrs.define(frozen=True, kw_only=True)
class PositionTrackingReward(Reward):
    """RewardState the closeness of the body to the target position more for the longer it has been doing so."""

    tracked_body_idx: int = attrs.field()
    base_body_idx: int = attrs.field()
    command_name: str = attrs.field()
    body_name: str = attrs.field()
    norm: xax.NormType = attrs.field(default="l1", validator=norm_validator)
    temp: float = attrs.field(default=1.0)
    monotonic_fn: MonotonicFn = attrs.field(default="inv")

    def get_reward(self, trajectory: Trajectory) -> Array:
        body_pos = trajectory.xpos[..., self.tracked_body_idx, :]
        base_pos = trajectory.xpos[..., self.base_body_idx, :]
        target_pos = trajectory.command[self.command_name][..., :3]
        error = xax.get_norm((body_pos - base_pos) - target_pos, self.norm).sum(-1)
        reward = norm_to_reward(error, self.temp, self.monotonic_fn)
        return reward

    @classmethod
    def create(
        cls,
        model: PhysicsModel,
        command_name: str,
        tracked_body_name: str,
        base_body_name: str,
        norm: xax.NormType = "l1",
        temp: float = 1.0,
        monotonic_fn: MonotonicFn = "inv",
        scale: float = 1.0,
        scale_by_curriculum: bool = False,
    ) -> Self:
        body_idx = get_body_data_idx_from_name(model, tracked_body_name)
        base_body_idx = get_body_data_idx_from_name(model, base_body_name)
        return cls(
            tracked_body_idx=body_idx,
            base_body_idx=base_body_idx,
            norm=norm,
            command_name=command_name,
            body_name=tracked_body_name,
            scale=scale,
            scale_by_curriculum=scale_by_curriculum,
            temp=temp,
            monotonic_fn=monotonic_fn,
        )

    def get_name(self) -> str:
        return f"{self.body_name}_{super().get_name()}"


@attrs.define(frozen=True, kw_only=True)
class UprightReward(Reward):
    """Reward for staying upright."""

    def get_reward(self, trajectory: Trajectory) -> Array:
        local_z = jnp.array([0.0, 0.0, 1.0])
        quat = trajectory.qpos[..., 3:7]
        global_z = xax.rotate_vector_by_quat(local_z, quat)
        return global_z[..., 2] - jnp.linalg.norm(global_z[..., :2], axis=-1)


@attrs.define(frozen=True, kw_only=True)
class LinkAccelerationPenalty(Reward):
    """Penalty for high link accelerations in the world frame."""

    norm: xax.NormType = attrs.field(default="l2", validator=norm_validator)

    def get_reward(self, trajectory: Trajectory) -> Array:
        pos = trajectory.xpos[..., 1:, :]
        pos_zp = jnp.pad(pos, ((2, 0), (0, 0), (0, 0)), mode="edge")
        done = jnp.pad(trajectory.done, ((2, 0),), mode="edge")[..., :-1, None]
        vel = jnp.where(done, 0.0, jnp.linalg.norm(pos_zp[..., 1:, :, :] - pos_zp[..., :-1, :, :], axis=-1))
        acc = jnp.where(done[..., 1:, :], 0.0, vel[..., 1:, :] - vel[..., :-1, :])
        penalty = xax.get_norm(acc, self.norm).mean(axis=-1)
        return penalty


@attrs.define(frozen=True, kw_only=True)
class LinkJerkPenalty(Reward):
    """Penalty for high link jerks in the world frame."""

    norm: xax.NormType = attrs.field(default="l2", validator=norm_validator)

    def get_reward(self, trajectory: Trajectory) -> Array:
        pos = trajectory.xpos[..., 1:, :]
        pos_zp = jnp.pad(pos, ((3, 0), (0, 0), (0, 0)), mode="edge")
        done = jnp.pad(trajectory.done, ((3, 0),), mode="edge")[..., :-1, None]
        vel = jnp.where(done, 0.0, jnp.linalg.norm(pos_zp[..., 1:, :, :] - pos_zp[..., :-1, :, :], axis=-1))
        acc = jnp.where(done[..., 1:, :], 0.0, vel[..., 1:, :] - vel[..., :-1, :])
        jerk = jnp.where(done[..., 2:, :], 0.0, acc[..., 1:, :] - acc[..., :-1, :])
        penalty = xax.get_norm(jerk, self.norm).mean(axis=-1)
        return penalty


@attrs.define(frozen=True, kw_only=True)
class SymmetryReward(Reward):
    """Rewards joints for having symmetrical positions."""

    joint_indices: tuple[int, ...] = attrs.field()
    joint_targets: tuple[float, ...] = attrs.field()

    def get_reward(self, trajectory: Trajectory) -> Array:
        qpos = trajectory.qpos[..., jnp.array(self.joint_indices) + 7] - jnp.array(self.joint_targets)
        qpos_mean = qpos.mean(axis=-2, keepdims=True)
        return (qpos * -qpos_mean).sum(axis=-1)

    @classmethod
    def create(
        cls,
        physics_model: PhysicsModel,
        joint_names: tuple[str, ...],
        joint_targets: tuple[float, ...],
        scale: float = 1.0,
        scale_by_curriculum: bool = False,
    ) -> Self:
        joint_to_idx = get_qpos_data_idxs_by_name(physics_model)
        joint_indices = tuple([int(joint_to_idx[name][0]) - 7 for name in joint_names])
        return cls(
            joint_indices=joint_indices,
            joint_targets=joint_targets,
            scale=scale,
            scale_by_curriculum=scale_by_curriculum,
        )


@attrs.define(frozen=True, kw_only=True)
class JoystickReward(Reward):
    """Reward for tracking the joystick commands.

    This reward uses global coordinates, so the robot should always start
    facing forward in the X direction.
    """

    forward_speed: float = attrs.field()
    backward_speed: float = attrs.field()
    strafe_speed: float = attrs.field()
    rotation_speed: float = attrs.field()
    command_name: str = attrs.field(default="joystick_command")
    lin_vel_penalty_scale: float = attrs.field(default=0.01)
    ang_vel_penalty_scale: float = attrs.field(default=0.01)
    norm: xax.NormType = attrs.field(default="l2", validator=norm_validator)
    temp: float = attrs.field(default=0.1)
    monotonic_fn: MonotonicFn = attrs.field(default="inv")
    in_robot_frame: bool = attrs.field(default=False)
    stand_base_reward: float = attrs.field(default=1.0)

    def get_reward(self, trajectory: Trajectory) -> Array:
        if self.command_name not in trajectory.command:
            raise ValueError(f"Command {self.command_name} not found! Ensure that it is in the task.")
        joystick_cmd = trajectory.command[self.command_name]
        chex.assert_shape(joystick_cmd, (..., 7))

        qvel = trajectory.qvel[..., :6]
        linvel = qvel[..., :3]
        angvel = qvel[..., 3:]

        if self.in_robot_frame:
            linvel = xax.rotate_vector_by_quat(linvel, trajectory.qpos[..., 3:7], inverse=True)

        # Penalty to discourage movement in general.
        linvel_norm = jnp.linalg.norm(linvel, axis=-1) * self.lin_vel_penalty_scale
        angvel_norm = jnp.linalg.norm(angvel, axis=-1) * self.ang_vel_penalty_scale
        vel_norm = linvel_norm + angvel_norm

        def normalize(x: Array, scale: float) -> Array:
            return x.clip(-scale, scale) / scale

        # Computes each of the penalties.
        stand_still_reward = self.stand_base_reward * jnp.ones_like(linvel[..., 0])
        walk_forward_reward = normalize(linvel[..., 0], self.forward_speed)
        walk_backward_reward = normalize(-linvel[..., 0], self.backward_speed)
        turn_left_reward = normalize(angvel[..., 2], self.rotation_speed)
        turn_right_reward = normalize(-angvel[..., 2], self.rotation_speed)
        strafe_left_reward = normalize(linvel[..., 1], self.strafe_speed)
        strafe_right_reward = normalize(-linvel[..., 1], self.strafe_speed)

        all_rewards = jnp.stack(
            [
                stand_still_reward,
                walk_forward_reward,
                walk_backward_reward,
                turn_left_reward,
                turn_right_reward,
                strafe_left_reward,
                strafe_right_reward,
            ],
            axis=-1,
        )

        # Weights each of the rewards by the one-hot encoded command.
        total_reward = jnp.einsum("...i,...i->...", joystick_cmd, all_rewards) - vel_norm

        return total_reward


@attrs.define(frozen=True, kw_only=True)
class ReachabilityPenalty(Reward):
    """Penalty for commands that exceed the per‑joint reachability envelope.

    Maximum joint movement in a single time step is given by:
    Δmax = vmax·Δt + ½·amax·Δt²
    """

    delta_max_j: tuple[float, ...] = attrs.field(eq=False, hash=False)
    squared: bool = True  # ‑‑ Use L2 on the excess (L1 if False).

    def get_reward(self, traj: Trajectory) -> jnp.ndarray:
        action_tj = traj.action
        q_prev_tj = traj.qpos[..., 7:]
        dm_j = jnp.asarray(self.delta_max_j)

        excess_tj = jnp.maximum(0.0, jnp.abs(action_tj - q_prev_tj) - dm_j)
        per_joint = excess_tj**2 if self.squared else excess_tj
        penalty_t = per_joint.sum(axis=-1)

        return penalty_t



"""Defines a teacher mixin for tasks."""

__all__ = [
    "StudentTask",
    "save_teacher",
    "StudentConfig",
    "TeacherReward",
]

import logging
import os
from abc import ABC, abstractmethod
from dataclasses import dataclass, replace
from pathlib import Path
from typing import Generic, TypeVar

import attrs
import distrax
import equinox as eqx
import jax
import jax.numpy as jnp
import xax
from jaxtyping import Array, PyTree

from ksim.rewards import Reward
from ksim.task.ppo import PPOConfig, PPOTask
from ksim.task.rl import (
    RolloutConstants,
    RolloutEnvState,
    RolloutSharedState,
)
from ksim.types import Trajectory

TEACHER_OUTPUT_KEY = "_teacher_output"

logger = logging.getLogger(__name__)


@jax.tree_util.register_dataclass
@dataclass
class StudentConfig(PPOConfig):
    """Configuration for teacher-student training."""

    # Teacher model path
    teacher_model_path: str = xax.field(
        value="",
        help="The path to the teacher model directory.",
    )


Config = TypeVar("Config", bound=StudentConfig)


@attrs.define(frozen=True, kw_only=True)
class TeacherReward(Reward):
    """Reward based on teacher output for teacher-student training."""

    def get_reward(self, trajectory: Trajectory) -> Array:
        if trajectory.aux_outputs is None or TEACHER_OUTPUT_KEY not in trajectory.aux_outputs:
            raise ValueError(
                "TeacherReward auxiliary output is missing! Make sure you are using it "
                "within the context of a teacher-student task, "
                "which populates the auxiliary output for you."
            )

        teacher_dist = trajectory.aux_outputs[TEACHER_OUTPUT_KEY]
        student_action = trajectory.action

        assert isinstance(teacher_dist, distrax.Distribution)

        log_prob = teacher_dist.log_prob(student_action)
        log_prob_max = teacher_dist.log_prob(teacher_dist.mode())  # value at mode
        reward = jnp.exp(log_prob - log_prob_max)  # 1.0 at the teacher mean
        return reward.mean(axis=-1)


class StudentTask(PPOTask[Config], Generic[Config], ABC):
    """Student task.

    This task extends PPO to include teacher-student training.
    """

    _teacher_policy: PyTree | None = None

    def run(self) -> None:
        """Load the teacher model before running the task."""
        self.load_teacher(Path(self.config.teacher_model_path), self.get_teacher_template())
        super().run()

    @abstractmethod
    def get_teacher_template(self) -> PyTree:
        """Get the teacher model template."""
        raise NotImplementedError("Subclasses must implement this method.")

    @abstractmethod
    def run_teacher(self, obs: xax.FrozenDict[str, Array], cmd: xax.FrozenDict[str, Array]) -> distrax.Distribution:
        """Run the teacher model.

        obs and cmd have shape (timesteps, ...)
        """
        raise NotImplementedError("Subclasses must implement this method.")

    def load_teacher(self, path: Path, like: PyTree) -> None:
        """Load the teacher model from a path."""
        self._teacher_policy = eqx.tree_deserialise_leaves(path / "teacher.eqx", like=like)

    @property
    def teacher_policy(self) -> PyTree:
        if self._teacher_policy is None:
            raise ValueError("Teacher model not loaded! Call load_teacher() first.")

        return self._teacher_policy

    def postprocess_trajectory(
        self,
        constants: RolloutConstants,
        env_states: RolloutEnvState,
        shared_state: RolloutSharedState,
        trajectory: Trajectory,
    ) -> Trajectory:
        trajectory = super().postprocess_trajectory(
            constants=constants,
            env_states=env_states,
            shared_state=shared_state,
            trajectory=trajectory,
        )

        action_dist = self.run_teacher(trajectory.obs, trajectory.command)

        # adds the teacher distribution to the aux outputs
        aux_outputs = trajectory.aux_outputs.unfreeze() if trajectory.aux_outputs else {}
        aux_outputs[TEACHER_OUTPUT_KEY] = action_dist
        trajectory = replace(trajectory, aux_outputs=xax.FrozenDict(aux_outputs))

        return trajectory


def save_teacher(policy: PyTree, path: Path) -> None:
    """Save the teacher model to a path."""
    os.makedirs(path, exist_ok=True)
    eqx.tree_serialise_leaves(path / "teacher.eqx", policy)



"""Defines the base actuators class, along with some implementations."""

__all__ = [
    "Actuators",
    "StatefulActuators",
    "TorqueActuators",
    "PositionActuators",
    "PositionVelocityActuator",
]

import logging
from abc import ABC, abstractmethod
from typing import Literal

import chex
import jax
import jax.numpy as jnp
from jaxtyping import Array, PRNGKeyArray, PyTree

from ksim.types import Metadata, PhysicsData, PhysicsModel
from ksim.utils.mujoco import get_ctrl_data_idx_by_name

logger = logging.getLogger(__name__)

NoiseType = Literal["none", "uniform", "gaussian"]


class Actuators(ABC):
    """Collection of actuators."""

    @classmethod
    def add_noise(cls, noise: float, noise_type: NoiseType, action: Array, rng: PRNGKeyArray) -> Array:
        match noise_type:
            case "none":
                return action
            case "uniform":
                return action + (jax.random.uniform(rng, action.shape) * 2 - 1) * noise
            case "gaussian":
                return action + jax.random.normal(rng, action.shape) * noise
            case _:
                raise ValueError(f"Invalid noise type: {noise_type}")

    @abstractmethod
    def get_ctrl(self, action: Array, physics_data: PhysicsData, rng: PRNGKeyArray) -> Array:
        """Get the control signal from the action vector."""

    def get_default_action(self, physics_data: PhysicsData) -> Array:
        """Get the default action for the actuators."""
        return physics_data.ctrl


class StatefulActuators(Actuators):
    @abstractmethod
    def get_stateful_ctrl(
        self,
        action: Array,
        physics_data: PhysicsData,
        actuator_state: PyTree,
        rng: PRNGKeyArray,
    ) -> tuple[Array, PyTree]:
        """Get the control signal from the action vector."""

    @abstractmethod
    def get_initial_state(self, physics_data: PhysicsData, rng: PRNGKeyArray) -> PyTree:
        """Get the initial state for the actuator."""

    def get_ctrl(self, action: Array, physics_data: PhysicsData, rng: PRNGKeyArray) -> Array:
        raise NotImplementedError("Stateful actuators should use `get_stateful_ctrl` instead.")


class TorqueActuators(Actuators):
    """Direct torque control."""

    def __init__(self, noise: float = 0.0, noise_type: NoiseType = "none") -> None:
        super().__init__()

        self.noise = noise
        self.noise_type = noise_type

    def get_ctrl(self, action: Array, physics_data: PhysicsData, rng: PRNGKeyArray) -> Array:
        """Just use the action as the torque, the simplest actuator model."""
        return self.add_noise(self.noise, self.noise_type, action, rng)


class PositionActuators(Actuators):
    """MIT Cheetah-style actuator controller operating on position."""

    def __init__(
        self,
        physics_model: PhysicsModel,
        metadata: Metadata,
        action_noise: float = 0.0,
        action_noise_type: NoiseType = "none",
        torque_noise: float = 0.0,
        torque_noise_type: NoiseType = "none",
    ) -> None:
        """Creates easily vector multipliable kps and kds."""
        ctrl_name_to_idx = get_ctrl_data_idx_by_name(physics_model)
        kps_list = [-1.0] * len(ctrl_name_to_idx)
        kds_list = [-1.0] * len(ctrl_name_to_idx)
        ctrl_clip_list = [jnp.inf] * len(ctrl_name_to_idx)

        if metadata.joint_name_to_metadata is None:
            raise ValueError("Joint metadata is required for MITPositionActuators")
        joint_name_to_metadata = metadata.joint_name_to_metadata

        for joint_name, params in joint_name_to_metadata.items():
            actuator_name = self.get_actuator_name(joint_name)
            if actuator_name not in ctrl_name_to_idx:
                if actuator_name != "root":
                    logger.warning("Joint %s has no actuator name. Skipping.", joint_name)
                continue
            actuator_idx = ctrl_name_to_idx[actuator_name]

            kp = params.kp
            kd = params.kd
            ctrl_clip = params.soft_torque_limit
            assert kp is not None and kd is not None, f"Missing kp or kd for joint {joint_name}"

            if ctrl_clip is not None:
                if ctrl_clip < 0:
                    raise ValueError(f"Soft torque limit for joint {joint_name} is negative: {ctrl_clip}")
                ctrl_clip_list[actuator_idx] = ctrl_clip

            kps_list[actuator_idx] = kp
            kds_list[actuator_idx] = kd

        if any(kp == -1 for kp in kps_list):
            raise ValueError("Some KPs are not set. Check the provided metadata.")
        if any(kd == -1 for kd in kds_list):
            raise ValueError("Some KDs are not set. Check the provided metadata.")

        self.kps = jnp.array(kps_list)
        self.kds = jnp.array(kds_list)
        self.ctrl_clip = jnp.array(ctrl_clip_list)

        self.action_noise = action_noise
        self.action_noise_type = action_noise_type
        self.torque_noise = torque_noise
        self.torque_noise_type = torque_noise_type

        if any(self.kps < 0) or any(self.kds < 0):
            raise ValueError("Some KPs or KDs are negative. Check the provided metadata.")
        if any(self.kps == 0) or any(self.kds == 0):
            logger.warning("Some KPs or KDs are 0. Check the provided metadata.")

    def get_actuator_name(self, joint_name: str) -> str:
        # This can be overridden if necessary.
        return f"{joint_name}_ctrl"

    def get_ctrl(self, action: Array, physics_data: PhysicsData, rng: PRNGKeyArray) -> Array:
        """Get the control signal from the (position) action vector."""
        pos_rng, tor_rng = jax.random.split(rng)
        current_pos = physics_data.qpos[7:]  # First 7 are always root pos.
        current_vel = physics_data.qvel[6:]  # First 6 are always root vel.
        target_velocities = jnp.zeros_like(action)
        pos_delta = self.add_noise(self.action_noise, self.action_noise_type, action - current_pos, pos_rng)
        vel_delta = target_velocities - current_vel

        ctrl = self.kps * pos_delta + self.kds * vel_delta

        return jnp.clip(
            self.add_noise(self.torque_noise, self.torque_noise_type, ctrl, tor_rng),
            -self.ctrl_clip,
            self.ctrl_clip,
        )


class PositionVelocityActuator(PositionActuators):
    """MIT Cheetah-style actuator controller operating on both position and velocity."""

    def __init__(
        self,
        physics_model: PhysicsModel,
        metadata: Metadata,
        pos_action_noise: float = 0.0,
        pos_action_noise_type: NoiseType = "none",
        vel_action_noise: float = 0.0,
        vel_action_noise_type: NoiseType = "none",
        torque_noise: float = 0.0,
        torque_noise_type: NoiseType = "none",
    ) -> None:
        super().__init__(
            physics_model=physics_model,
            metadata=metadata,
            action_noise=pos_action_noise,
            action_noise_type=pos_action_noise_type,
            torque_noise=torque_noise,
            torque_noise_type=torque_noise_type,
        )

        self.vel_action_noise = vel_action_noise
        self.vel_action_noise_type = vel_action_noise_type

    def get_ctrl(self, action: Array, physics_data: PhysicsData, rng: PRNGKeyArray) -> Array:
        """Get the control signal from the (position and velocity) action vector."""
        pos_rng, vel_rng, tor_rng = jax.random.split(rng, 3)

        current_pos = physics_data.qpos[7:]  # First 7 are always root pos.
        current_vel = physics_data.qvel[6:]  # First 6 are always root vel.

        # Adds position and velocity noise.
        target_position = action[: len(current_pos)]
        target_velocity = action[len(current_pos) :]
        chex.assert_equal_shape([current_pos, target_position, current_vel, target_velocity])
        target_position = self.add_noise(self.action_noise, self.action_noise_type, target_position, pos_rng)
        target_velocity = self.add_noise(self.vel_action_noise, self.vel_action_noise_type, target_velocity, vel_rng)

        pos_delta = target_position - current_pos
        vel_delta = target_velocity - current_vel

        ctrl = self.kps * pos_delta + self.kds * vel_delta
        return jnp.clip(
            self.add_noise(self.torque_noise, self.torque_noise_type, ctrl, tor_rng),
            -self.ctrl_clip,
            self.ctrl_clip,
        )

    def get_default_action(self, physics_data: PhysicsData) -> Array:
        """Get the default action (zeros) with the correct shape."""
        qpos_dim = len(physics_data.qpos[7:])
        return jnp.zeros(qpos_dim * 2)
